@@ -2,7 +2,7 @@
 """Ingest Generic Database tables using multi-value insert statements and multiprocessing.
 
 Usage:
-  %s <configFile> <inputFile>... [--fileoffiles] [--table=<table>] [--tableDelimiter=<tableDelimiter>] [--bundlesize=<bundlesize>] [--nprocesses=<nprocesses>] [--nfileprocesses=<nfileprocesses>] [--loglocationInsert=<loglocationInsert>] [--logprefixInsert=<logprefixInsert>] [--loglocationIngest=<loglocationIngest>] [--logprefixIngest=<logprefixIngest>] [--types=<types>] [--skiphtm] [--nullValue=<nullValue>] [--fktable=<fktable>] [--fktablecols=<fktablecols>] [--fktablecoltypes=<fktablecoltypes>] [--fkfield=<fkfield>] [--fkfrominputdata=<fkfrominputdata>] [--racol=<racol>] [--deccol=<deccol>]
+  %s <configFile> <inputFile>... [--fileoffiles] [--table=<table>] [--tableDelimiter=<tableDelimiter>] [--bundlesize=<bundlesize>] [--nprocesses=<nprocesses>] [--nfileprocesses=<nfileprocesses>] [--loglocationInsert=<loglocationInsert>] [--logprefixInsert=<logprefixInsert>] [--loglocationIngest=<loglocationIngest>] [--logprefixIngest=<logprefixIngest>] [--columns=<columns>] [--types=<types>] [--skiphtm] [--nullValue=<nullValue>] [--fktable=<fktable>] [--fktablecols=<fktablecols>] [--fktablecoltypes=<fktablecoltypes>] [--fkfield=<fkfield>] [--fkfrominputdata=<fkfrominputdata>] [--racol=<racol>] [--deccol=<deccol>]
   %s (-h | --help)
   %s --version
 
@@ -19,8 +19,9 @@ Options:
   --logprefixInsert=<logprefixInsert>      Log prefix [default: inserter]
   --loglocationIngest=<loglocationIngest>  Log file location [default: /tmp/]
   --logprefixIngest=<logprefixIngest>      Log prefix [default: ingester]
-  --types=<types>                          Python column types in the same order as the column headers.
-  --skiphtm                                Don't bother calculating HTMs. They're either already done or we don't need them.
+  --columns=<columns>                      List of columns, comma separated, no spaces. If blank, assumes all columns of the input data.
+  --types=<types>                          PYTHON column types in the same order as the column headers.
+  --skiphtm                                Don't bother calculating HTMs. They're either already done or we don't need them. (I.e. not spatially indexed data.)
   --nullValue=<nullValue>                  Value of NULL definition (e.g. NaN, NULL, \\N, None) [default: \\N]
   --fktable=<fktable>                      Cassandra has a flat schema, so join to another table file via a foreign key (e.g. exposures).
   --fktablecols=<fktablecols>              The valid columns in the foreign key table we want to use - comma separated, no spaces (e.g. expname,object,mjd,filter,mag5sig,zp_mag,fwhm_px,exptime,detem).
@@ -34,23 +35,80 @@ Example:
   %s config_cassandra.yaml 01a58464o0535o.dph --fktable=/Users/kws/atlas/dophot/all_co_exposures.tst --fkfield=expname --fktablecols=mjd,expname,exptime,filter,mag5sig --types=float,float,float,int,int,float,float,float,float,float,float,float,float,float,float,float,float,float --fktablecoltypes=float,str,float,str,float --table=atlasdophot --racol=RA --deccol=Dec
 
   %s /home/kws/config_cassandra_atlas.yaml /home/kws/atlas/dophot/ingest/parallel_machine_ingest_test/remaining_batch/exposures_around_galactic_centre_10degrees_20210219_cleaned_hko_only_second_attempt_db1 --fileoffiles --fktable=/home/kws/atlas/dophot/all_co_exposures.tst --fkfield=expname --fktablecols=mjd,expname,exptime,filter,mag5sig --types=float,float,float,int,int,float,float,float,float,float,float,float,float,float,float,float,float,float --fktablecoltypes=float,str,float,str,float --table=atlas_detections --racol=RA --deccol=Dec --nprocesses=8 --nfileprocesses=4 --loglocationIngest=/home/kws/cassandra_ingest_logs/db1/cassandra_ingest/galactic_centre_hko/ --loglocationInsert=/home/kws/cassandra_ingest_logs/db1/cassandra_ingest/galactic_centre_hko/
+
+  %s /Users/kws/config_cassandra.yaml /Users/kws/lasair/cassandra/load-old-data/noncandidates/file_of_files_to_ingest.txt --fileoffiles --types=str,float,float,int,int,float,float,float,int --table=test_noncandidates --tableDelimiter=, --nprocesses=11 --nfileprocesses=1 --skiphtm
 """
 import sys
-__doc__ = __doc__ % (sys.argv[0], sys.argv[0], sys.argv[0], sys.argv[0], sys.argv[0])
+__doc__ = __doc__ % (sys.argv[0], sys.argv[0], sys.argv[0], sys.argv[0], sys.argv[0], sys.argv[0])
 from docopt import docopt
 import os, shutil, re
 from gkutils.commonutils import Struct, cleanOptions, readGenericDataFile, dbConnect, which, splitList, parallelProcess
 from datetime import datetime
 from datetime import timedelta
 import subprocess
-#import MySQLdb
 from cassandra.cluster import Cluster
 import gzip
 from collections import OrderedDict
 
 # 2021-02-11 KWS Import the new htmNameBulk function! No need anymore to rely on an external binary!
 #                No need to write temporary files anymore.
-from gkhtm._gkhtm import htmNameBulk
+from gkhtm._gkhtm import htmNameBulk, htmIDBulk
+
+def readZTFAvroPacket(filename, addhtm16 = None):
+    from fastavro import reader
+    print(filename)
+    with open(filename, 'rb') as f:
+        candlist = []
+        nondetectionCandlist = []
+        avro_reader = reader(f)
+        for record in avro_reader:
+            prv_candidates = []
+            try:
+                if record['prv_candidates'] is not None:
+                    prv_candidates = record['prv_candidates']
+            except KeyError as e:
+                pass
+            candidates = [record['candidate']] + prv_candidates
+            for cand in candidates:
+   
+                # Remove the images.
+                try:
+                    del cand['cutoutDifference']
+                except KeyError as e:
+                    pass
+                try:
+                    del cand['cutoutTemplate']
+                except KeyError as e:
+                    pass
+                try:
+                    del cand['cutoutScience']
+                except KeyError as e:
+                    pass
+
+                cand['objectId'] = record['objectId']
+
+                if not 'candid' in cand or not cand['candid']:
+                    nondetectionCandlist.append({'objectId': cand['objectId'],
+                                                 'jd': cand['jd'],
+                                                 'fid': cand['fid'],
+                                                 'diffmaglim': cand['diffmaglim'],
+                                                 'nid': cand['nid'],
+                                                 'field': cand['field'],
+                                                 'magzpsci': cand['magzpsci'],
+                                                 'magzpsciunc': cand['magzpsciunc'],
+                                                 'magzpscirms': cand['magzpscirms']})
+                else:
+                    candlist.append(cand)
+
+    if addhtm16 is not None:
+        coords = [[x['ra'], x['dec']] for x in candlist]
+        htm16s = htmIDBulk(16, coords)
+        for i in range(len(candlist)):
+            candlist[i]['htm16'] = htm16s[i]
+
+    data = {'candidates': candlist, 'noncandidates': nondetectionCandlist}
+
+    return data
 
 
 def nullValue(value, nullValue = '\\N'):
@@ -77,46 +135,38 @@ def boolToInteger(value):
         returnValue = 0
     return returnValue
 
-def calculate_htm_name_bulk(generateHtmNameBulk, htmLevel, tempRaDecFile):
-    # Call the C++ HTM ID calculator
-    htmIDs = []
-
-    p = subprocess.Popen([generateHtmNameBulk, str(htmLevel), tempRaDecFile], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
-    output, errors = p.communicate()
-
-    if output.strip():
-        htmIDs = output.strip().split('\n')
-
-    return htmIDs
-
 
 # Use INSERT statements so we can use multiprocessing
-def executeLoad(session, table, data, bundlesize = 100, types = None):
+# 2021-10-16 KWS Why do we need types?? This is because if we send the data as a CSV dict,
+#                then Cassandra will not know what to do with the data. A float != string.
+#                MySQL is a bit more generous inasmuch as it will auto cast. Not Cassandra.
+#                We allow this option so that we can pass the data directly from CSV.
+#                An alternative approach is to modify readGenericDataFile so that it will
+#                cast during the load. Avro dictionaries are already typed.
+def executeLoad(session, table, data, bundlesize = 1, types = None):
 
     rowsUpdated = 0
 
     if len(data) == 0:
+        print('No data!')
         return rowsUpdated
 
-    if types is None:
-        return rowsUpdated
+    #if types is None:
+    #    return rowsUpdated
 
     keys = list(data[0].keys())
 
-
-    if len(keys) != len(types):
-        print("Keys & Types mismatch")
-        return rowsUpdated
-
     typesDict = OrderedDict()
 
-    i = 0
-    for k in keys:
-        typesDict[k] = types[i]
-        i += 1
+    if types is not None:
+        if len(keys) != len(types):
+            print("Keys & Types mismatch")
+            return rowsUpdated
+        i = 0
+        for k in keys:
+            typesDict[k] = types[i]
+            i += 1
 
-    #print(keys)
-    #print(types)
 
     formatSpecifier = ','.join(['%s' for i in keys])
 
@@ -138,14 +188,23 @@ def executeLoad(session, table, data, bundlesize = 100, types = None):
             sql += ';'
 
             values = []
-            for row in dataChunk:
-                for key in keys:
-                    value = nullValueNULL(boolToInteger(row[key]))
-                    if value is not None:
-                        value = eval(typesDict[key])(value)
-                    values.append(value)
 
-            #print(sql)
+            for row in dataChunk:
+                # If data comes from a CSV. We need to cast the results using the types. Otherwise assume
+                # the types are already correct. (E.g. data read from an Avro file.)
+                for key in keys:
+                    if types is not None:
+                        value = nullValueNULL(boolToInteger(row[key]))
+                        if value is not None:
+                            value = eval(typesDict[key])(value)
+                        values.append(value)
+                    # The data is already in the right python type. (Actually it doesn't matter! All the values are strings!)
+                    else:
+                        value = row[key]
+                        values.append(value)
+
+
+            print(sql, tuple(values))
             session.execute(sql, tuple(values))
 
 
@@ -153,7 +212,6 @@ def executeLoad(session, table, data, bundlesize = 100, types = None):
             template = "An exception of type {0} occurred. Arguments:\n{1!r}"
             message = template.format(type(e).__name__, e.args)
             print(message)
-            #print "Error %d: %s" % (e.args[0], e.args[1])
 
     return
 
@@ -175,11 +233,12 @@ def workerInsert(num, db, objectListFragment, dateAndTime, firstPass, miscParame
 
     # Add 3 string columns if the HTMs are being requested. You will not be able to insert into a table
     # if its htm name components are not specified.
-    if not options.skiphtm:
+    if not options.skiphtm and combinedTypes is not None:
         combinedTypes = combinedTypes + ",str,str,str"
 
-    types = combinedTypes.split(',')
-    #print (types)
+    types = None
+    if combinedTypes is not None:
+        types = combinedTypes.split(',')
 
     # This is in the worker function
     objectsForUpdate = executeLoad(session, options.table, objectListFragment, int(options.bundlesize), types=types)
@@ -192,12 +251,6 @@ def workerInsert(num, db, objectListFragment, dateAndTime, firstPass, miscParame
 
 def ingestData(options, inputFiles, fkDict = None):
 
-    #if not options.skiphtm:
-    #    generateHtmNameBulk = which('generate_htmname_bulk')
-    #    if generateHtmNameBulk is None:
-    #        sys.stderr.write("Can't find the generate_htmname_bulk executable, so cannot continue.\n")
-    #        exit(1)
-
     import yaml
     with open(options.configFile) as yaml_file:
         config = yaml.safe_load(yaml_file)
@@ -224,14 +277,37 @@ def ingestData(options, inputFiles, fkDict = None):
 
     for inputFile in inputFiles:
         print("Ingesting %s" % inputFile)
-        if 'gz' in inputFile:
+        if '.gz' in inputFile:
             # It's probably gzipped
             f = gzip.open(inputFile, 'rb')
             print(type(f).__name__)
         else:
             f = inputFile
     
-        data = readGenericDataFile(f, delimiter=delimiter, useOrderedDict=True)
+        if 'avro' in inputFile:
+            # Data is in Avro packets, with schema. Let's hard-wire to the ZTF schema for the time being.
+            avroData = readZTFAvroPacket(f, addhtm16 = True)
+            if 'noncandidates' in options.table:
+                data = avroData['noncandidates']
+            elif 'candidates' in options.table:
+                data = avroData['candidates']
+            else:
+                print("Error. Incorrect table definition for Avro packets. Must contain candidates or noncandidates.")
+                exit(1)
+
+        else:
+            # Data is in plain text file. No schema present, so will need to provide
+            # column types.
+            data = readGenericDataFile(f, delimiter=delimiter, useOrderedDict=True)
+
+        # 2021-07-29 KWS This is a bit inefficient, but trim the data down to specified columns if they are present.
+        if options.columns:
+            trimmedData = []
+            for row in data:
+                trimmedRow = {key: row[key] for key in options.columns.split(',')}
+                trimmedData.append(trimmedRow)
+            data = trimmedData
+
 
         foreignKey = options.fkfrominputdata
         if foreignKey == 'filename':
@@ -253,24 +329,16 @@ def ingestData(options, inputFiles, fkDict = None):
                 except KeyError as e:
                     pass
 
-        print(data[0])
+        #print(data[0])
         pid = os.getpid()
     
         if not options.skiphtm:
-            #tempRADecFile = '/tmp/' + os.path.basename(inputFile) + 'radec_' + str(pid)
-            #tempLoadFile = '/tmp/' + os.path.basename(inputFile) + '_' + str(pid) + '.csv'
     
             coords = []
-            #with open(tempRADecFile, 'wb') as f:
-            #    for row in data:
-            #        f.write('%s %s\n' % (row['ra'], row['dec']))
             for row in data:
                 coords.append([float(row[options.racol]), float(row[options.deccol])])
     
-            #htm16Names = calculate_htm_name_bulk(generateHtmNameBulk, 16, tempRADecFile)
             htm16Names = htmNameBulk(16, coords)
-    
-            #os.remove(tempRADecFile)
 
             # For Cassandra, we're going to split the HTM Name across several columns.
             # Furthermore, we only need to do this once for the deepest HTM level, because
